
% FHCGMTT: fast hierarchical clustering based on growing multi-layer
% topology training. 4 neurons are trained for each sub-map with
% membership control (each node contains sqrt(n) data points or less)

function [HAcc, FM] = GMTTMain(X,NormType,LinkType,Pm)
tic
%% preprocessing
% initialize parameters
XLable=X(:,Pm.Xwd+1);
X=X(:,1:Pm.Xwd);
Pm.LearnRate1=0.1; % learn rate for the following steps
Pm.LearnRate0=0.01; % learn rate for the only one node in layer 0
Pm.ErrorBound=0.01; % learn rate for the following layers
Pm.B=4;
Pm.UpLimit=floor(sqrt(Pm.Xlth));

%normalization
if NormType==1 % max min normalization
    for i=1:Pm.Xwd
        MinV=min(X(:,i));
        MaxV=max(X(:,i));
        X(:,i)=X(:,i)-MinV;
        X(:,i) =X(:,i)/(MaxV-MinV);
    end
end
if NormType==2 % standard mean normalization
    for i=1:Pm.Xwd
        X(:,i)=(X(:,i)-mean(X(:,i)))./std(X(:,i));
    end
end

% train a node for layer 1
Node{1}=X(randperm(Pm.Xlth,1),:);
for i=1:Pm.Xlth
    Node{1}=Node{1}+Pm.LearnRate0*(X(i,:)-Node{1});
end

% pre-assign some values
Layer(1,1)=1; % map the layer No. and the sequential No. of each nodes
Layer(2,1)=2; 
Child{1}=1; % record if have child node for each nodes. 1: has child, 0: no child.
CorSet{1}=true(1,Pm.Xlth); % at the beginning, corset of the first node is the whole data set
StopGrow=0; % determiner to decide if to grow the topology or not
PreNode=1; % sequential number of the just processed node
ThisCoarse=1; % sequential number of the coarse node in the just processed layer
ThisLayer=2; % sequential number of the layer to be created
ThisNode=2; % sequential number of the node to be created
Counter=2; % No. of existing node
Reader=(1:Pm.Xlth); % reader to read out the sequential number of data objects

%% growing training
% recursively grow and train the topology
while ThisLayer~=1
    if StopGrow==0 % 0: grow deeper layer, 1: stop grow
        SubDataLct=Reader(CorSet{PreNode}(ThisCoarse,:)); % extract the "SubData" for deeper layer growth
        SubData=X(SubDataLct,:);
        %PrePreNode=find(Layer==ThisLayer-2, 1, 'last' );
        %if PrePreNode>0
            %PreCoarse=length(find(Layer==ThisLayer-2));
            %NodeGrandParents=Node{PrePreNode}(PreCoarse,:);
        %else
            %NodeGrandParents=NaN;
        %end
        %[Node{ThisNode},CorSet{ThisNode},NumNode] = ...
            %TrainNodes(SubData,SubDataLct,Pm,Node{PreNode}(ThisCoarse,:),Node{PreNode},NodeGrandParents); % ***function*** %
        [Node{ThisNode},CorSet{ThisNode},NumNode] = ...
            TrainNodes(SubData,SubDataLct,Pm);
        Child{ThisNode}=zeros(1,NumNode);
    end
    Bad1=find(sum(CorSet{ThisNode},2)>Pm.UpLimit); % find the node with high membership
    Bad2=find(Child{ThisNode}==0); % find the node with no heir
    CoarsePool=intersect(Bad1,Bad2);
    if isempty(CoarsePool)==1 % back to check the upper layer
        ThisLayer=ThisLayer-1;
        ThisNode=PreNode;
        PreNode=find(Layer==ThisLayer-1, 1, 'last' );
        StopGrow=1;
    else % if nodes have no heir and too many data points
        ThisCoarse=CoarsePool(1);
        Counter=Counter+1;
        Child{ThisNode}(ThisCoarse)=1;
        PreNode=ThisNode;
        ThisNode=Counter;
        ThisLayer=ThisLayer+1;
        Layer(ThisNode,1)=ThisLayer;
        StopGrow=0;
    end
end
Pm.TotalNumNode=Counter;
toc
Sigma=mean(pdist(X))/10;
%% form "Links" through the trained GHSOM
[Links] = MergeLayer(Layer,Node,CorSet,LinkType,Sigma); % ***function*** %

%% calculate the hierarchy accuracy and FM index
FM = Eva_FM(Pm,Child,Layer,CorSet,Links,XLable); % ***function*** %
HAcc = Eva_HAcc(X,Pm,Child,CorSet,Links,XLable,LinkType);
end




