
% FHCGMTT: fast hierarchical clustering based on growing multi-layer
% topology training. 4 neurons are trained for each sub-map with
% membership control (each node contains sqrt(n) data points or less)

function HAcc = IGMTTMain(X,NormType,PartRatio,Pm)
tic
%% preprocessing
% initialize parameters
%[Pm.Xlth,Pm.Xwd]=size(X);
%RandList=randperm(Pm.Xlth,Pm.Xlth);
%X=X(RandList,:);
XLable=X(:,Pm.Xwd+1);
%Pm.NumClass=length(unique(XLable));
%Pm.Xwd=Pm.Xwd-1;
X=X(:,1:Pm.Xwd);
Pm.LearnRate1=0.1; % learn rate for the following steps
Pm.LearnRate0=0.01; % learn rate for the only one node in layer 0
Pm.ErrorBound=0.01; % learn rate for the following layers
Pm.B=4;
Pm.UpLimit=floor(sqrt(Pm.Xlth));
Pm.P=PartRatio;
Pm.UpLimitP=floor(sqrt(Pm.Xlth*Pm.P));
Pm.XlthP=round(Pm.Xlth*Pm.P);

%normalization
if NormType==1 % max min normalization
    for i=1:Pm.Xwd
        MinV=min(X(:,i));
        MaxV=max(X(:,i));
        X(:,i)=X(:,i)-MinV;
        X(:,i) =X(:,i)/(MaxV-MinV);
    end
end
if NormType==2 % standard mean normalization
    for i=1:Pm.Xwd
        X(:,i)=(X(:,i)-mean(X(:,i)))./std(X(:,i));
    end
end

% train a coarse topology using CGMTT
[Node,Child,CorSet,Layer]=CGMTTMain(X,Pm);
Counter=length(Layer);
Reader=(1:Pm.Xlth); % reader to read out the sequential number of data objects
for i=Pm.XlthP+1:Pm.Xlth
    FoundIt=0;
    ThisLayer=1;
    NodeLct=1;
    Winner=1;
    while FoundIt==0      
        CorSet{NodeLct}(Winner,i)=1;
        [~,Winner]=min(dist(X(i,:),Node{NodeLct}'));
        if Child{NodeLct}(Winner)==0
            FoundIt=1;
        else    
            LayerRank=length(find(Child{NodeLct}(1:Winner)==1));
            ThisLayer=ThisLayer+1;
            LayerList=find(Layer==ThisLayer);
            ct=0;
            for j=1:length(LayerList)
                if LayerList(j)>NodeLct
                    ct=ct+1;
                end
                if ct==LayerRank
                    NodeLct=LayerList(j);
                end
            end
        end
    end
    if length(find(CorSet{NodeLct}(Winner,:)==1))>Pm.UpLimit*(Pm.B-1) % full-coarse node detected, should train
        Counter=Counter+1;
        SubDataLct=Reader(CorSet{NodeLct}(Winner,:)); % extract the "SubData" for deeper layer growth
        SubData=X(SubDataLct,:);
        [Node{Counter},CorSet{Counter},NumNode] = ...
            ITrainNodes(SubData,SubDataLct,Pm); % ***function*** %
        Child{NodeLct}(Winner)=1;
        Child{Counter}=zeros(1,NumNode);
        Layer(Counter)=ThisLayer+1;
    end
end

% pre-assign some values
StopGrow=0; % determiner to decide if to grow the topology or not
PreNode=1; % sequential number of the just processed node
ThisCoarse=1; % sequential number of the coarse node in the just processed layer
ThisLayer=2; % sequential number of the layer to be created
ThisNode=2; % sequential number of the node to be created
Counter=2; % No. of existing node
Reader=(1:Pm.Xlth); % reader to read out the sequential number of data objects

%% growing training
% recursively grow and train the topology
while ThisLayer~=1
    if StopGrow==0 % 0: grow deeper layer, 1: stop grow
        SubDataLct=Reader(CorSet{PreNode}(ThisCoarse,:)); % extract the "SubData" for deeper layer growth
        SubData=X(SubDataLct,:);
        [Node{ThisNode},CorSet{ThisNode},NumNode] = ...
            ITrainNodes(SubData,SubDataLct,Pm); % ***function*** %
        Child{ThisNode}=zeros(1,NumNode);
    end
    Bad1=find(sum(CorSet{ThisNode},2)>Pm.UpLimit); % find the node with high membership
    Bad2=find(Child{ThisNode}==0); % find the node with no heir
    CoarsePool=intersect(Bad1,Bad2);
    if isempty(CoarsePool)==1 % back to check the upper layer
        ThisLayer=ThisLayer-1;
        ThisNode=PreNode;
        PreNode=find(Layer==ThisLayer-1, 1, 'last' );
        StopGrow=1;
    else % if nodes have no heir and too many data points
        ThisCoarse=CoarsePool(1);
        Counter=Counter+1;
        Child{ThisNode}(ThisCoarse)=1;
        PreNode=ThisNode;
        ThisNode=Counter;
        ThisLayer=ThisLayer+1;
        Layer(ThisNode,1)=ThisLayer;
        StopGrow=0;
    end
end
toc
Pm.TotalNumNode=Counter;
Sigma=mean(pdist(X))/10;
%% form "Links" through the trained GHSOM
[Links] = IMergeLayer(Layer,Node,CorSet,4,Sigma); % ***function*** %

%% calculate the hierarchy accuracy
HAcc = Eva_HAcc(X,Pm,Child,CorSet,Links,XLable,4);

end




